#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains functions for constructing CoolProp expressions.
"""
from pyomo.environ import exp, units as pyunits, Var
from pyomo.core.expr.calculus.derivatives import Modes, differentiate

from idaes.core.util.exceptions import ConfigurationError


# TODO : Only have temperature derivative expression for exponential_tau form
# TODO : Add other derivative forms as/if required


def parameters_nt_sum(cobj, prop, nlist, tlist):
    """
    Create parameters for expression forms using n-t parameters

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        nlist: list of values for n-parameter
        tlist: list of values for t-parameter

    Returns:
        None
    """
    if len(nlist) != len(tlist):
        raise ConfigurationError(
            f"{cobj.name} mismatched length between n and t parameters "
            f"for CoolProp exponential form for property {prop}. Please "
            f"ensure the number of n and t parameters are equal."
        )

    # Use multiple Vars, instead of single indexed Var, to have same
    # structure as cases where each parameter value has different units

    for i, nval in enumerate(nlist):
        coeff = Var(
            doc="Multiplying parameter for CoolProp exponential form",
            units=pyunits.dimensionless,
        )
        cobj.add_component(prop + "_coeff_n" + str(i + 1), coeff)
        coeff.fix(nval)

    for i, tval in enumerate(tlist):
        coeff = Var(
            doc="Exponent parameter for CoolProp exponential form",
            units=pyunits.dimensionless,
        )
        cobj.add_component(prop + "_coeff_t" + str(i + 1), coeff)
        coeff.fix(tval)


def _nt_sum(cobj, prop, theta):
    """
    Create sum expressions in n-t forms (sum(n[i]*theta**t[i]))

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        theta: expression or variable to use for theta in expression

    Returns:
        Pyomo expression of sum term
    """
    # Build sum term
    i = 1
    s = 0
    while True:
        try:
            ni = getattr(cobj, f"{prop}_coeff_n{i}")
            ti = getattr(cobj, f"{prop}_coeff_t{i}")
            s += ni * theta**ti
            i += 1
        except AttributeError:
            break
    return s


def expression_exponential(cobj, prop, T, yc, tau=False):
    """
    Create expressions for CoolProp exponential sum forms. This function
    supports both exponential forms used by CoolProp:

    Without tau: y = yc * exp(sum(ni*theta^ti))
    With tau: y = yc * exp((Tc/T) * sum(ni*theta^ti))

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        T: temperature to use in expression
        yc: value of property at critical point
        tau: whether tau=Tc/T should be included in expression (default=False)

    Returns:
        Pyomo expression matching CoolProp exponential sum form
    """
    Tc = cobj.temperature_crit
    theta = 1 - T / Tc

    s = _nt_sum(cobj, prop, theta)

    if tau:
        return yc * exp(Tc / T * s)
    else:
        return yc * exp(s)


def dT_expression_exponential(cobj, prop, T, yc, tau=False):
    """
    Create expressions for temperature derivative of CoolProp exponential sum
    forms with tau

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        T: temperature to use in expression
        yc: value of property at critical point
        tau: whether tau=Tc/T should be included in expression (default=False)

    Returns:
        Pyomo expression for temperature derivative of CoolProp exponential sum
        form with tau
    """
    # y = yc * exp(Tc/T * sum(ni*theta^ti))
    # Need d(y)/dT

    y = expression_exponential(cobj, prop, T, yc, tau)

    return differentiate(expr=y, wrt=T, mode=Modes.reverse_symbolic)


def expression_nonexponential(cobj, prop, T, yc):
    """
    Create expressions for CoolProp non-exponential sum forms

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        T: temperature to use in expression
        yc: value of property at critical point

    Returns:
        Pyomo expression matching CoolProp non-exponential sum form
    """
    # y = yc * (1 + sum(ni*theta^ti))
    Tc = cobj.temperature_crit
    theta = 1 - T / Tc

    s = _nt_sum(cobj, prop, theta)

    return yc * (1 + s)


def parameters_polynomial(cobj, prop, prop_units, alist, blist):
    """
    Create parameters for expression forms using A-B parameters (rational
    polynomial forms)

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        prop_units: units of measurement for property
        Alist: list of values for A-parameter
        Blist: list of values for B-parameter

    Returns:
        None
    """
    for i, aval in enumerate(alist):
        if i == 0:
            param_units = prop_units
        else:
            param_units = prop_units / pyunits.K**i

        coeff = Var(doc="A parameter for CoolProp polynomial form", units=param_units)
        cobj.add_component(prop + "_coeff_A" + str(i), coeff)
        coeff.fix(aval)

    for i, bval in enumerate(blist):
        if i == 0:
            param_units = pyunits.dimensionless
        else:
            param_units = pyunits.K**-i

        coeff = Var(doc="B parameter for CoolProp exponential form", units=param_units)
        cobj.add_component(prop + "_coeff_B" + str(i), coeff)
        coeff.fix(bval)


def expression_polynomial(cobj, prop, T):
    """
    Create expressions for CoolProp rational polynomial forms

    Args:
        cobj: Component object that will contain the parameters
        prop: name of property parameters are associated with
        T: temperature to use in expression

    Returns:
        Pyomo expression matching CoolProp rational polynomial form
    """
    i = 0
    asum = 0
    try:
        while True:
            Ai = getattr(cobj, f"{prop}_coeff_A{i}")
            asum += Ai * T**i
            i += 1
    except AttributeError:
        pass

    i = 0
    bsum = 0
    try:
        while True:
            Bi = getattr(cobj, f"{prop}_coeff_B{i}")
            bsum += Bi * T**i
            i += 1
    except AttributeError:
        pass

    return asum / bsum



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
This module contains the IDAES get_solver method.
"""

import idaes.logger as idaeslog
import idaes.core.solvers

_log = idaeslog.getLogger(__name__)


# Author: Andrew Lee
def get_solver(
    solver=None,
    solver_options: dict = None,
    writer_config: dict = None,
    options: dict = None,
):
    """
    General method for getting a solver object which defaults to the standard
    IDAES solver (defined in the IDAES configuration).

    Args:
        solver: string name for desired solver. Default=None, use default solver
        solver_options: dict of solver options to use, overwrites any settings
                 provided by IDAES configuration. Default = None, use default
                 solver options.
        writer_config: dict of configuration options for solver writer, overwrites
                 ny settings provided by IDAES configuration. Default = None, use
                 default solver options.
        options: DEPRECATED. Alias of solver_options.

    Returns:
        A Pyomo solver object
    """
    if solver_options is not None:
        if options is not None:
            raise ValueError(
                "Cannot provide both the 'options' and 'solver_options' argument. "
                "'options' has been deprecated in favor of 'solver_options'."
            )
        options = solver_options

    if solver is None:
        solver = "default"
    solver_obj = idaes.core.solvers.SolverWrapper(solver, register=False)()

    if options is not None:
        for k, v in options.items():
            solver_obj.options[k] = v
    if writer_config is not None:
        for k, v in writer_config.items():
            solver_obj.config.writer_config[k] = v

    return solver_obj



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Waterwall section model

main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction and gravity
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag


"""
# Import Pyomo libraries
from pyomo.environ import (
    value,
    Var,
    Param,
    asin,
    cos,
    sqrt,
    log10,
    PositiveReals,
    Reference,
    units as pyunits,
)
from pyomo.dae import DerivativeVar
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.core.expr import Expr_if

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)

from idaes.core.util.config import is_physical_parameter_block, DefaultBool
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog


# Additional import for the unit operation
from idaes.core.util.constants import Constants as const


__author__ = "Boiler Subsystem Team (J. Ma, M. Zamarripa)"
__version__ = "2.0.0"


@declare_process_block_class("WaterwallSection")
class WaterwallSectionData(UnitModelBlockData):
    """
    WaterwallSection Unit Class
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=DefaultBool,
            default=useDefault,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentPhase,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of material balance should be constructed,
**default** - MaterialBalanceType.componentPhase.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single ethalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - ethalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "rigorous_boiling",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Heat of reaction term construction flag",
            doc="""Indicates whether terms for heat of reaction terms should be
constructed,
**default** - False.
**Valid values:** {
**True** - include heat of reaction terms,
**False** - exclude heat of reaction terms.}""",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation)


        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super().build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_geometry()
        # This model requires the IAPWS95 property package with the mixed phase
        # option, therefore, phase equilibrium calculations are handled by
        # the property package.
        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add Ports
        self.add_inlet_port()
        self.add_outlet_port()

        # Add object references
        self.volume = Reference(self.control_volume.volume)

        # Set references to balance terms at unit level
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat)

        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != "none"
        ):
            self.deltaP = Reference(self.control_volume.deltaP)

        # Set Unit Geometry and Holdup Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary, and link to holdup volume
        """
        units_meta = self.config.property_package.get_metadata()
        # Total projected wall area of waterwall section from fire-side model
        self.projected_area = Var(
            initialize=10,
            doc="Total projected wall area of waterwall section",
            units=units_meta.get_derived_units("area"),
        )
        # Number of waterwall tubes
        self.number_tubes = Var(initialize=4, doc="Number of waterwall tubes")
        # Height of waterwall section, given by boiler model
        self.height = Var(
            initialize=5.0,
            doc="Height of waterwall section",
            units=units_meta.get_derived_units("length"),
        )
        # Length of waterwall tubes calculated based on given total area
        # and perimeter of waterwall
        self.tube_length = Var(
            initialize=5.0,
            doc="length waterwall tube",
            units=units_meta.get_derived_units("length"),
        )
        # Inner diameter of waterwall tubes
        self.tube_diameter = Var(
            initialize=0.05,
            doc="Inside diameter of waterwall tube",
            units=units_meta.get_derived_units("length"),
        )

        # Inside radius of waterwall tube
        @self.Expression(doc="Inside radius of waterwall tube")
        def radius_inner(b):
            return 0.5 * b.tube_diameter

        # Total cross section area of fluid flow
        @self.Expression(doc="Cross section area of fluid")
        def area_cross_fluid_total(b):
            return 0.25 * const.pi * b.tube_diameter**2 * b.number_tubes

        # Tube thickness
        self.tube_thickness = Var(
            initialize=0.005,
            doc="Thickness of waterwall tube",
            units=units_meta.get_derived_units("length"),
        )

        # Outside radius of waterwall tube
        @self.Expression(doc="Outside radius of waterwall tube")
        def radius_outer(b):
            return b.radius_inner + b.tube_thickness

        # Thickness of waterwall fin
        self.fin_thickness = Var(
            initialize=0.004,
            doc="Thickness of waterwall Fin",
            units=units_meta.get_derived_units("length"),
        )

        # Half of the waterwall fin thickness
        @self.Expression(doc="Half of the waterwall fin thickness")
        def fin_thickness_half(b):
            return 0.5 * b.fin_thickness

        # Length of waterwall fin
        self.fin_length = Var(
            initialize=0.005,
            doc="Length of waterwall fin",
            units=units_meta.get_derived_units("length"),
        )
        # Thickness of slag layer
        self.slag_thickness = Var(
            self.flowsheet().time,
            bounds=(0.0001, 0.009),
            initialize=0.001,
            doc="Thickness of slag layer",
            units=units_meta.get_derived_units("length"),
        )

        @self.Expression(doc="Pitch of two neighboring tubes")
        def pitch(b):
            return b.fin_length + b.radius_outer * 2.0

        # Equivalent tube length (not necessarily equal to height)
        @self.Constraint(doc="Equivalent length of tube")
        def tube_length_eqn(b):
            return b.tube_length * b.pitch * b.number_tubes == b.projected_area

        @self.Expression(doc="Angle at joint of tube and fin")
        def alpha_tube(b):
            return asin(b.fin_thickness_half / b.radius_outer)

        @self.Expression(
            self.flowsheet().time,
            doc="Angle at joint of tube and " "fin at outside slag layer",
        )
        def alpha_slag(b, t):
            return asin(
                (b.fin_thickness_half + b.slag_thickness[t])
                / (b.radius_outer + b.slag_thickness[t])
            )

        @self.Expression(doc="Perimeter of interface between slag and tube")
        def perimeter_interface(b):
            return (
                (const.pi - 2 * b.alpha_tube) * b.radius_outer
                + b.pitch
                - 2 * b.radius_outer * cos(b.alpha_tube)
            )

        @self.Expression(doc="Perimeter on the inner tube side")
        def perimeter_ts(b):
            return const.pi * b.tube_diameter

        @self.Expression(self.flowsheet().time, doc="Perimeter on the outer slag side")
        def perimeter_ss(b, t):
            return (
                (const.pi - 2 * b.alpha_slag[t])
                * (b.radius_outer + b.slag_thickness[t])
                + b.pitch
                - 2 * (b.radius_outer + b.slag_thickness[t]) * cos(b.alpha_slag[t])
            )

        # Cross section area of tube and fin metal
        @self.Expression(doc="Cross section area of tube and fin metal")
        def area_cross_metal(b):
            return (
                const.pi * (b.radius_outer**2 - b.radius_inner**2)
                + b.fin_thickness * b.fin_length
            )

        # Cross section area of slag layer
        @self.Expression(
            self.flowsheet().time, doc="Cross section area of slag layer per tube"
        )
        def area_cross_slag(b, t):
            return b.perimeter_interface * b.slag_thickness[t]

        # Volume constraint
        @self.Constraint(
            self.flowsheet().time, doc="waterwall fluid volume of all tubes"
        )
        def volume_eqn(b, t):
            return (
                b.volume[t]
                == 0.25 * const.pi * b.tube_diameter**2 * b.tube_length * b.number_tubes
            )

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.
        """
        units_meta = self.config.property_package.get_metadata()
        self.fcorrection_dp = Var(
            initialize=1.2,
            within=PositiveReals,
            doc="correction factor for pressure drop due to acceleration"
            "and unsmooth tube applied to friction term",
        )
        # Thermal conductivity of metal
        self.therm_cond_metal = Param(
            initialize=43.0,
            mutable=True,
            doc="Thermal conductivity of tube metal",
            units=units_meta.get_derived_units("thermal_conductivity"),
        )
        # Thermal conductivity of slag
        self.therm_cond_slag = Param(
            initialize=1.3,
            mutable=True,
            doc="Thermal conductivity of slag",
            units=units_meta.get_derived_units("thermal_conductivity"),
        )
        # Heat capacity of metal
        self.cp_metal = Param(
            initialize=500.0,
            mutable=True,
            doc="Heat capacity of tube metal",
            units=units_meta.get_derived_units("heat_capacity_mass"),
        )
        # Heat Capacity of slag
        self.cp_slag = Param(
            initialize=250,
            mutable=True,
            doc="Heat capacity of slag",
            units=units_meta.get_derived_units("heat_capacity_mass"),
        )
        # Density of metal
        self.dens_metal = Param(
            initialize=7800.0,
            mutable=True,
            doc="Density of tube metal",
            units=units_meta.get_derived_units("density_mass"),
        )
        # Density of slag
        self.dens_slag = Param(
            initialize=2550,
            mutable=True,
            doc="Density of slag",
            units=units_meta.get_derived_units("density_mass"),
        )
        # Shape factor of tube metal conduction resistance
        # based on projected area
        self.fshape_metal = Param(
            initialize=0.7718, mutable=True, doc="Shape factor of tube metal conduction"
        )
        # Shape factor of slag conduction resistance
        # based on projected area
        self.fshape_slag = Param(
            initialize=0.6858, mutable=True, doc="Shape factor of slag conduction"
        )
        # Shape factor of convection on tube side resistance
        # based on projected area
        self.fshape_conv = Param(
            initialize=0.8496, mutable=True, doc="Shape factor of convection"
        )

        # Heat conduction resistance of half of metal wall thickness
        # based on interface perimeter
        @self.Expression(doc="half metal layer conduction resistance")
        def half_resistance_metal(b):
            return (
                b.tube_thickness
                / 2
                / b.therm_cond_metal
                * b.fshape_metal
                * b.perimeter_interface
                / b.pitch
            )

        # Heat conduction resistance of half of slag thickness
        # based on mid slag layer perimeter
        @self.Expression(
            self.flowsheet().time, doc="half slag layer conduction resistance"
        )
        def half_resistance_slag(b, t):
            return (
                b.slag_thickness[t]
                / 2
                / b.therm_cond_slag
                * b.fshape_slag
                * (b.perimeter_ss[t] + b.perimeter_interface)
                / 2
                / b.pitch
            )

        # Add performance variables
        # Heat from fire side boiler model
        self.heat_fireside = Var(
            self.flowsheet().time,
            initialize=1e7,
            doc="total heat from fire side model for the section",
            units=units_meta.get_derived_units("power"),
        )
        # Tube boundary wall temperature
        self.temp_tube_boundary = Var(
            self.flowsheet().time,
            initialize=400.0,
            doc="Temperature of tube boundary wall",
            units=units_meta.get_derived_units("temperature"),
        )
        # Tube center point wall temperature
        self.temp_tube_center = Var(
            self.flowsheet().time,
            initialize=450.0,
            doc="Temperature of tube center wall",
            units=units_meta.get_derived_units("temperature"),
        )
        # Slag boundary wall temperature
        self.temp_slag_boundary = Var(
            self.flowsheet().time,
            initialize=600.0,
            doc="Temperature of slag boundary wall",
            units=units_meta.get_derived_units("temperature"),
        )
        # Slag center point wall temperature
        self.temp_slag_center = Var(
            self.flowsheet().time,
            initialize=500.0,
            doc="Temperature of slag layer center point",
            units=units_meta.get_derived_units("temperature"),
        )

        # Energy holdup for slag layer
        self.energy_holdup_slag = Var(
            self.flowsheet().time,
            initialize=1e4,
            doc="Energy holdup of slag layer",
            units=(
                units_meta.get_derived_units("energy")
                * units_meta.get_derived_units("length") ** -1
            ),
        )

        # Energy holdup for metal (tube + fin)
        self.energy_holdup_metal = Var(
            self.flowsheet().time,
            initialize=1e6,
            doc="Energy holdup of metal",
            units=(
                units_meta.get_derived_units("energy")
                * units_meta.get_derived_units("length") ** -1
            ),
        )

        # Energy accumulation for slag and metal
        if self.config.dynamic is True:
            self.energy_accumulation_slag = DerivativeVar(
                self.energy_holdup_slag,
                wrt=self.flowsheet().time,
                doc="Energy accumulation of slag layer",
            )
            self.energy_accumulation_metal = DerivativeVar(
                self.energy_holdup_metal,
                wrt=self.flowsheet().time,
                doc="Energy accumulation of tube and fin metal",
            )

        def energy_accumulation_term_slag(b, t):
            return b.energy_accumulation_slag[t] if b.config.dynamic else 0

        def energy_accumulation_term_metal(b, t):
            return b.energy_accumulation_metal[t] if b.config.dynamic else 0

        # Velocity of liquid only
        self.velocity_liquid = Var(
            self.flowsheet().time,
            initialize=3.0,
            doc="Velocity of liquid only",
            units=units_meta.get_derived_units("velocity"),
        )

        # Reynolds number based on liquid only flow
        self.N_Re = Var(self.flowsheet().time, initialize=1.0e6, doc="Reynolds number")

        # Prandtl number of liquid phase
        self.N_Pr = Var(self.flowsheet().time, initialize=2.0, doc="Reynolds number")

        # Darcy friction factor
        self.friction_factor_darcy = Var(
            self.flowsheet().time, initialize=0.01, doc="Darcy friction factor"
        )

        # Vapor fraction at inlet
        self.vapor_fraction = Var(
            self.flowsheet().time,
            initialize=0.0,
            doc="Vapor fractoin of vapor-liquid mixture",
        )

        # Liquid fraction at inlet
        self.liquid_fraction = Var(
            self.flowsheet().time,
            initialize=1.0,
            doc="Liquid fractoin of vapor-liquid mixture",
        )

        # Density ratio of liquid to vapor
        self.ratio_density = Var(
            self.flowsheet().time, initialize=1.0, doc="Liquid to vapor density ratio"
        )

        # Void fraction at inlet
        self.void_fraction = Var(
            self.flowsheet().time, initialize=0.0, doc="void fraction at inlet"
        )

        # Exponent n for gamma at inlet, typical range in (0.75,0.8294)
        self.n_exp = Var(
            self.flowsheet().time, initialize=0.82, doc="exponent for gamma at inlet"
        )

        # Gamma for velocity slip at inlet
        self.gamma = Var(
            self.flowsheet().time,
            initialize=3.0,
            doc="gamma for velocity slip at inlet",
        )

        # Mass flux
        self.mass_flux = Var(
            self.flowsheet().time,
            initialize=1000.0,
            doc="mass flux",
            units=units_meta.get_derived_units("flux_mass"),
        )

        # Reduced pressure
        self.reduced_pressure = Var(
            self.flowsheet().time, initialize=0.85, doc="reduced pressure"
        )

        # Two-phase correction factor
        self.phi_correction = Var(
            self.flowsheet().time,
            initialize=1.01,
            doc="Two-phase flow correction factor",
        )

        # Convective heat transfer coefficient on tube side,
        # typically in range (1000, 5e5)
        self.hconv = Var(
            self.flowsheet().time,
            initialize=30000.0,
            doc="Convective heat transfer coefficient",
            units=units_meta.get_derived_units("heat_transfer_coefficient"),
        )

        # Convective heat transfer coefficient for liquid only,
        # typically in range (1000.0, 1e5)
        self.hconv_liquid = Var(
            self.flowsheet().time,
            initialize=20000.0,
            doc="Convective heat transfer coefficient of liquid only",
            units=units_meta.get_derived_units("heat_transfer_coefficient"),
        )

        # Pool boiling heat transfer coefficient, typically in range (1e4, 5e5)
        self.hpool = Var(
            self.flowsheet().time,
            initialize=1e5,
            doc="Pool boiling heat transfer coefficient",
            units=units_meta.get_derived_units("heat_transfer_coefficient"),
        )

        # Boiling number, typical range in (1e-7, 5e-4) in original formula.
        # we define here as boiling_number_scaled == 1e6*boiling_number
        self.boiling_number_scaled = Var(
            self.flowsheet().time, initialize=1, doc="Scaled boiling number"
        )

        # Enhancement factor, typical range in (1.0, 3.0)
        self.enhancement_factor = Var(
            self.flowsheet().time, initialize=1.3, doc="Enhancement factor"
        )

        # Suppression factor, typical range in (0.005, 1.0)
        self.suppression_factor = Var(
            self.flowsheet().time, initialize=0.03, doc="Suppression factor"
        )

        # Convective heat flux to fluid
        self.heat_flux_conv = Var(
            self.flowsheet().time,
            initialize=7e4,
            doc="Convective heat flux to fluid",
            units=units_meta.get_derived_units("flux_energy"),
        )

        # Slag-tube interface heat flux
        self.heat_flux_interface = Var(
            self.flowsheet().time,
            initialize=100000.0,
            doc="Slag-tube interface heat flux",
            units=units_meta.get_derived_units("flux_energy"),
        )

        # Pressure change due to friction
        self.deltaP_friction = Var(
            self.flowsheet().time,
            initialize=-1000.0,
            doc="Pressure change due to friction",
            units=units_meta.get_derived_units("pressure"),
        )

        # Pressure change due to gravity
        self.deltaP_gravity = Var(
            self.flowsheet().time,
            initialize=-1000.0,
            doc="Pressure change due to gravity",
            units=units_meta.get_derived_units("pressure"),
        )

        # Equation to calculate heat flux to slag boundary
        @self.Expression(self.flowsheet().time, doc="heat flux at slag outer layer")
        def heat_flux_fireside(b, t):
            return b.heat_fireside[t] * b.pitch / (b.projected_area * b.perimeter_ss[t])

        # Equation to calculate slag layer boundary temperature
        @self.Constraint(self.flowsheet().time, doc="slag layer boundary temperature")
        def slag_layer_boundary_temperature_eqn(b, t):
            return b.heat_flux_fireside[t] * b.half_resistance_slag[t] == (
                b.temp_slag_boundary[t] - b.temp_slag_center[t]
            )

        # Equation to calculate heat flux at the slag-metal interface
        @self.Constraint(self.flowsheet().time, doc="heat flux at slag-tube interface")
        def heat_flux_interface_eqn(b, t):
            return b.heat_flux_interface[t] * (
                b.half_resistance_metal + b.half_resistance_slag[t]
            ) == (b.temp_slag_center[t] - b.temp_tube_center[t])

        # Equation to calculate heat flux at tube boundary
        @self.Constraint(
            self.flowsheet().time, doc="convective heat flux at tube boundary"
        )
        def heat_flux_conv_eqn(b, t):
            return b.heat_flux_conv[
                t
            ] * b.fshape_conv * b.perimeter_ts == b.pitch * b.hconv[t] * (
                b.temp_tube_boundary[t] - b.control_volume.properties_in[t].temperature
            )

        # Equation to calculate tube boundary wall temperature
        @self.Constraint(self.flowsheet().time, doc="tube boundary wall temperature")
        def temperature_tube_boundary_eqn(b, t):
            return b.heat_flux_conv[
                t
            ] * b.half_resistance_metal * b.perimeter_ts == b.perimeter_interface * (
                b.temp_tube_center[t] - b.temp_tube_boundary[t]
            )

        # Equation to calculate energy holdup for slag layer per tube length
        @self.Constraint(self.flowsheet().time, doc="energy holdup for slag layer")
        def energy_holdup_slag_eqn(b, t):
            return (
                b.energy_holdup_slag[t]
                == b.temp_slag_center[t]
                * b.cp_slag
                * b.dens_slag
                * b.area_cross_slag[t]
            )

        # Equation to calculate energy holdup for metal
        # (tube + fin) per tube length
        @self.Constraint(self.flowsheet().time, doc="energy holdup for metal")
        def energy_holdup_metal_eqn(b, t):
            return (
                b.energy_holdup_metal[t]
                == b.temp_tube_center[t]
                * b.cp_metal
                * b.dens_metal
                * b.area_cross_metal
            )

        # Energy balance for slag layer
        @self.Constraint(self.flowsheet().time, doc="energy balance for slag layer")
        def energy_balance_slag_eqn(b, t):
            return (
                energy_accumulation_term_slag(b, t)
                == b.heat_flux_fireside[t] * b.perimeter_ss[t]
                - b.heat_flux_interface[t] * b.perimeter_interface
            )

        # Energy balance for metal
        @self.Constraint(self.flowsheet().time, doc="energy balance for metal")
        def energy_balance_metal_eqn(b, t):
            return (
                energy_accumulation_term_metal(b, t)
                == b.heat_flux_interface[t] * b.perimeter_interface
                - b.heat_flux_conv[t] * b.perimeter_ts
            )

        # Expression to calculate slag/tube metal interface wall temperature
        @self.Expression(
            self.flowsheet().time, doc="Slag tube interface wall temperature"
        )
        def temp_interface(b, t):
            return (
                b.temp_tube_center[t]
                + b.heat_flux_interface[t] * b.half_resistance_metal
            )

        # Equations for calculate pressure drop
        # and convective heat transfer coefficient for 2-phase flow
        # Equation to calculate liquid to vapor density ratio
        @self.Constraint(self.flowsheet().time, doc="liquid to vapor density ratio")
        def ratio_density_eqn(b, t):
            return (
                0.01
                * b.ratio_density[t]
                * b.control_volume.properties_in[t].dens_mol_phase["Vap"]
                == 0.01 * b.control_volume.properties_in[t].dens_mol_phase["Liq"]
            )

        # Equation for calculating velocity if the flow is liquid only
        @self.Constraint(self.flowsheet().time, doc="Vecolity of fluid if liquid only")
        def velocity_lo_eqn(b, t):
            return (
                1e-4
                * b.velocity_liquid[t]
                * b.area_cross_fluid_total
                * b.control_volume.properties_in[t].dens_mol_phase["Liq"]
                == 1e-4 * b.control_volume.properties_in[t].flow_mol
            )

        # Equation for calculating Reynolds number if liquid only
        @self.Constraint(self.flowsheet().time, doc="Reynolds number if liquid only")
        def Reynolds_number_eqn(b, t):
            return (
                b.N_Re[t] * b.control_volume.properties_in[t].visc_d_phase["Liq"]
                == b.tube_diameter
                * b.velocity_liquid[t]
                * b.control_volume.properties_in[t].dens_mass_phase["Liq"]
            )

        # Friction factor depending on laminar or turbulent flow,
        # usually always turbulent (>1187.385)
        @self.Constraint(self.flowsheet().time, doc="Darcy friction factor")
        def friction_factor_darcy_eqn(b, t):
            return b.friction_factor_darcy[t] * b.N_Re[t] ** 0.25 / 0.3164 == 1.0

        # Vapor fractoin equation at inlet,
        # add 1e-5 such that vapor fraction is always positive
        @self.Constraint(self.flowsheet().time, doc="Vapor fractoin at inlet")
        def vapor_fraction_eqn(b, t):
            return 100 * b.vapor_fraction[t] == 100 * (
                b.control_volume.properties_in[t].vapor_frac + 1e-5
            )

        # n-exponent equation for inlet
        @self.Constraint(self.flowsheet().time, doc="n-exponent")
        def n_exp_eqn(b, t):
            return 0.001 * (0.8294 - b.n_exp[t]) * b.control_volume.properties_in[
                t
            ].pressure == 8.0478 * units_meta.get_derived_units("pressure")

        # Gamma equation for inlet
        @self.Constraint(self.flowsheet().time, doc="Gamma at inlet")
        def gamma_eqn(b, t):
            return b.gamma[t] == b.ratio_density[t] ** b.n_exp[t]

        # void faction at inlet equation
        @self.Constraint(self.flowsheet().time, doc="Void fractoin at inlet")
        def void_fraction_eqn(b, t):
            return (
                b.void_fraction[t] * (1.0 + b.vapor_fraction[t] * (b.gamma[t] - 1.0))
                == b.vapor_fraction[t] * b.gamma[t]
            )

        # Two-phase flow correction factor equation
        @self.Constraint(self.flowsheet().time, doc="Correction factor")
        def correction_factor_eqn(b, t):
            return (b.phi_correction[t] - 0.027 * b.liquid_fraction[t]) ** 2 == (
                0.973 * b.liquid_fraction[t] + b.vapor_fraction[t] * b.ratio_density[t]
            ) * (0.973 * b.liquid_fraction[t] + b.vapor_fraction[t])

        # Pressure change equation due to friction,
        # -1/2*density*velocity^2*fD/diameter*length*phi^2
        @self.Constraint(self.flowsheet().time, doc="pressure change due to friction")
        def pressure_change_friction_eqn(b, t):
            return (
                0.01 * b.deltaP_friction[t] * b.tube_diameter
                == -0.01
                * b.fcorrection_dp
                * 0.5
                * b.control_volume.properties_in[t].dens_mass_phase["Liq"]
                * b.velocity_liquid[t] ** 2
                * b.friction_factor_darcy[t]
                * b.tube_length
                * b.phi_correction[t] ** 2
            )

        # Pressure change equation due to gravity,
        # density_mixture*gravity*height
        @self.Constraint(self.flowsheet().time, doc="pressure change due to gravity")
        def pressure_change_gravity_eqn(b, t):
            return 1e-3 * b.deltaP_gravity[
                t
            ] == -1e-3 * const.acceleration_gravity * b.height * (
                b.control_volume.properties_in[t].dens_mass_phase["Vap"]
                * b.void_fraction[t]
                + b.control_volume.properties_in[t].dens_mass_phase["Liq"]
                * (1.0 - b.void_fraction[t])
            )

        # Mass flux of vapor-liquid mixture
        # (density*velocity or mass_flow/area)
        @self.Constraint(self.flowsheet().time, doc="mass flux")
        def mass_flux_eqn(b, t):
            return (
                b.mass_flux[t] * b.area_cross_fluid_total
                == b.control_volume.properties_in[t].flow_mol
                * b.control_volume.properties_in[0].mw
            )

        # Liquid fraction at inlet
        @self.Constraint(self.flowsheet().time, doc="liquid fraction")
        def liquid_fraction_eqn(b, t):
            return b.liquid_fraction[t] + b.vapor_fraction[t] == 1.0

        # Total pressure change equation
        @self.Constraint(self.flowsheet().time, doc="pressure drop")
        def pressure_change_total_eqn(b, t):
            return b.deltaP[t] == b.deltaP_friction[t] + b.deltaP_gravity[t]

        # Total heat added to control_volume
        @self.Constraint(
            self.flowsheet().time, doc="total heat added to fluid control_volume"
        )
        def heat_eqn(b, t):
            return (
                b.heat_duty[t]
                == b.number_tubes * b.heat_flux_conv[t] * b.tube_length * b.perimeter_ts
            )

        # Reduced pressure
        @self.Constraint(self.flowsheet().time, doc="reduced pressure")
        def reduced_pressure_eqn(b, t):
            return (
                b.reduced_pressure[t] * self.config.property_package.pressure_crit
                == b.control_volume.properties_in[t].pressure
            )

        # Prandtl number of liquid
        @self.Constraint(self.flowsheet().time, doc="liquid Prandtl number")
        def N_Pr_eqn(b, t):
            return (
                b.N_Pr[t]
                * b.control_volume.properties_in[t].therm_cond_phase["Liq"]
                * b.control_volume.properties_in[0].mw
                == b.control_volume.properties_in[t].cp_mol_phase["Liq"]
                * b.control_volume.properties_in[t].visc_d_phase["Liq"]
            )

        # Forced convection heat transfer coefficient for liquid only
        @self.Constraint(
            self.flowsheet().time,
            doc="forced convection heat transfer " "coefficient for liquid only",
        )
        def hconv_lo_eqn(b, t):
            return (
                b.hconv_liquid[t] * b.tube_diameter
                == 0.023
                * b.N_Re[t] ** 0.8
                * b.N_Pr[t] ** 0.4
                * b.control_volume.properties_in[t].therm_cond_phase["Liq"]
            )

        # Pool boiling heat transfer coefficient
        @self.Constraint(
            self.flowsheet().time, doc="pool boiling heat transfer coefficient"
        )
        def hpool_eqn(b, t):
            return (
                1e-4
                * b.hpool[t]
                * sqrt(
                    pyunits.convert(
                        b.control_volume.properties_in[0].mw,
                        to_units=pyunits.g / pyunits.mol,
                    )
                )
                * (-log10(b.reduced_pressure[t])) ** (0.55)
                == 1e-4
                * 55.0
                * b.reduced_pressure[t] ** 0.12
                * b.heat_flux_conv[t] ** 0.67
            )

        # Boiling number scaled by a factor of 1e6
        @self.Constraint(self.flowsheet().time, doc="boiling number")
        def boiling_number_eqn(b, t):
            return (
                1e-10
                * b.boiling_number_scaled[t]
                * b.control_volume.properties_in[t].dh_vap_mol
                * b.mass_flux[t]
                == b.heat_flux_conv[t] * b.control_volume.properties_in[0].mw * 1e-4
            )

        if self.config.rigorous_boiling is True:
            # Due to low contribution to the enhancement factor and highly nonlinear
            # constraint, the Martinelli parameter has been removed from the model.
            # if required user needs to declare the variable and constraint, and
            # update constraint in line 909 to add the Martinelli parameter factor

            # Reciprocal of Martinelli parameter to the power of 0.86,
            # typical range in (1e-3, 1.0)
            self.martinelli_reciprocal_p86 = Var(
                self.flowsheet().time,
                initialize=0.2,
                doc="Reciprocal of Martinelli parameter " "to the power of 0.86",
            )

            # Reciprocal of Martinelli parameter to the power of 0.86
            @self.Constraint(
                self.flowsheet().time,
                doc="Reciprocal of Martineli parameter " "to the power of 0.86",
            )
            def martinelli_reciprocal_p86_eqn(b, t):
                return b.martinelli_reciprocal_p86[t] * b.liquid_fraction[
                    t
                ] ** 0.774 * b.control_volume.properties_in[t].visc_d_phase[
                    "Liq"
                ] ** 0.086 == Expr_if(
                    b.vapor_fraction[t] > 0,
                    (b.vapor_fraction[t]) ** 0.774
                    * b.ratio_density[t] ** 0.43
                    * b.control_volume.properties_in[t].visc_d_phase["Vap"] ** 0.086,
                    0.0,
                )

        # Enhancement factor
        # due to low contribution the reciprocal Martinalli parameter,
        # it can be removed from the enhancement factor equation
        @self.Constraint(
            self.flowsheet().time, doc="Forced convection enhancement factor"
        )
        def enhancement_factor_eqn(b, t):
            if self.config.rigorous_boiling is True:
                return b.enhancement_factor[t] == 1.0 + 24000.0 * (
                    b.boiling_number_scaled[t] / 1e6
                ) ** 1.16 + Expr_if(
                    b.vapor_fraction[t] > 0, 1.37 * b.martinelli_reciprocal_p86[t], 0.0
                )
            else:
                return (
                    b.enhancement_factor[t]
                    == 1.0 + 24000.0 * (b.boiling_number_scaled[t] / 1e6) ** 1.16
                )

        # Suppression factor
        @self.Constraint(self.flowsheet().time, doc="Pool boiler suppression factor")
        def suppression_factor_eqn(b, t):
            return (
                b.suppression_factor[t]
                * (1.0 + 1.15e-6 * b.enhancement_factor[t] ** 2 * b.N_Re[t] ** 1.17)
                == 1.0
            )

        @self.Constraint(
            self.flowsheet().time, doc="convective heat transfer coefficient"
        )
        def hconv_eqn(b, t):
            return (
                1e-3 * b.hconv[t]
                == 1e-3 * b.hconv_liquid[t] * b.enhancement_factor[t]
                + 1e-3 * b.hpool[t] * b.suppression_factor[t]
            )

    def set_initial_condition(self):
        """Initialization of dynamic accumulation terms"""

        if self.config.dynamic is True:
            self.control_volume.material_accumulation[:, :, :].value = 0
            self.control_volume.energy_accumulation[:, :].value = 0
            self.control_volume.material_accumulation[0, :, :].fix(0)
            self.control_volume.energy_accumulation[0, :].fix(0)
            self.energy_accumulation_slag[:].value = 0
            self.energy_accumulation_metal[:].value = 0
            self.energy_accumulation_slag[0].fix(0)
            self.energy_accumulation_metal[0].fix(0)

    def initialize_build(
        blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        Waterwall section initialization routine.

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                           package(s) for the control_volume of the model to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        flags = blk.control_volume.initialize(
            outlvl=outlvl + 1, optarg=optarg, solver=solver, state_args=state_args
        )
        init_log.info_high("Initialization Step 1 Complete.")
        # Fix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].enth_mol.fix(
                value(blk.control_volume.properties_in[t].enth_mol)
                + value(blk.heat_fireside[t])
                / value(blk.control_volume.properties_in[t].flow_mol)
            )
            blk.control_volume.properties_out[t].pressure.fix(
                value(blk.control_volume.properties_in[t].pressure) - 1.0
            )
        blk.heat_eqn.deactivate()
        blk.pressure_change_total_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # Unfix outlet enthalpy and pressure
        for t in blk.flowsheet().time:
            blk.control_volume.properties_out[t].enth_mol.unfix()
            blk.control_volume.properties_out[t].pressure.unfix()
        blk.heat_eqn.activate()
        blk.pressure_change_total_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        blk.control_volume.release_state(flags, outlvl + 1)
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for t, c in self.energy_holdup_slag_eqn.items():
            s = iscale.get_scaling_factor(
                self.energy_holdup_slag[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.friction_factor_darcy_eqn.items():
            s = iscale.get_scaling_factor(self.N_Re[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.volume_eqn.items():
            s = iscale.get_scaling_factor(self.volume[t], default=1, warning=True)
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.heat_flux_conv_eqn.items():
            s = iscale.get_scaling_factor(
                self.heat_flux_conv[t], default=1, warning=True
            )
            s *= iscale.get_scaling_factor(self.tube_diameter, default=1, warning=True)
            iscale.constraint_scaling_transform(c, s / 10.0, overwrite=False)
        for t, c in self.energy_holdup_slag_eqn.items():
            s = iscale.get_scaling_factor(
                self.energy_holdup_slag[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)
        for t, c in self.energy_holdup_metal_eqn.items():
            s = iscale.get_scaling_factor(
                self.energy_holdup_metal[t], default=1, warning=True
            )
            iscale.constraint_scaling_transform(c, s, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for 0D Boiler heat exchanger model.

Author: Miguel Zamarripa
"""
import pytest

from pyomo.environ import (
    ConcreteModel,
    check_optimal_termination,
    value,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock

from idaes.models.properties import iapws95

# import ideal flue gas prop pack
from idaes.models_extra.power_generation.properties import (
    FlueGasParameterBlock,
)

# Import Power Plant HX Unit Model
from idaes.models_extra.power_generation.unit_models.boiler_heat_exchanger import (
    BoilerHeatExchanger,
    TubeArrangement,
    delta_temperature_lmtd_callback,
    delta_temperature_amtd_callback,
    delta_temperature_underwood_callback,
    HeatExchangerFlowPattern,
)

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.models.properties.general_helmholtz import helmholtz_available

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def tc(delta_temperature_callback=delta_temperature_underwood_callback):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(
        delta_temperature_callback=delta_temperature_callback,
        cold_side={"property_package": m.fs.prop_steam, "has_pressure_change": True},
        hot_side={"property_package": m.fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=True,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Liq",
        has_radiation=True,
    )

    # Check unit config arguments
    assert not m.fs.unit.config.dynamic
    assert m.fs.unit.config.has_holdup
    assert m.fs.unit.config.flow_pattern == HeatExchangerFlowPattern.countercurrent


def th(
    delta_temperature_callback=delta_temperature_underwood_callback,
    tout_1=809.55,
    tout_2=788.53,
):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(
        delta_temperature_callback=delta_temperature_callback,
        cold_side={"property_package": m.fs.prop_steam, "has_pressure_change": True},
        hot_side={"property_package": m.fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Liq",
        has_radiation=True,
    )

    #   Set inputs
    h = value(iapws95.htpx(773.15 * pyunits.K, 2.5449e7 * pyunits.Pa))
    m.fs.unit.cold_side_inlet.flow_mol[0].fix(24678.26)  # mol/s
    m.fs.unit.cold_side_inlet.enth_mol[0].fix(h)  # J/mol
    m.fs.unit.cold_side_inlet.pressure[0].fix(2.5449e7)  # Pascals

    # FLUE GAS Inlet from Primary Superheater
    FGrate = 28.3876e3 * 0.18  # mol/s equivalent of ~1930.08 klb/hr
    # Use FG molar composition to set component flow rates (baseline report)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "H2O"].fix(FGrate * 8.69 / 100)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "CO2"].fix(FGrate * 14.49 / 100)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "N2"].fix(FGrate * 74.34 / 100)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "O2"].fix(FGrate * 2.47 / 100)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "NO"].fix(FGrate * 0.0006)
    m.fs.unit.hot_side_inlet.flow_mol_comp[0, "SO2"].fix(FGrate * 0.002)
    m.fs.unit.hot_side_inlet.temperature[0].fix(1102.335)
    m.fs.unit.hot_side_inlet.pressure[0].fix(100145)

    # Primary Superheater
    ITM = 0.0254  # inch to meter conversion
    m.fs.unit.tube_di.fix((2.5 - 2 * 0.165) * ITM)
    m.fs.unit.tube_thickness.fix(0.165 * ITM)
    m.fs.unit.pitch_x.fix(3 * ITM)
    # gas path transverse width 54.78 ft / number of columns
    m.fs.unit.pitch_y.fix(54.78 / 108 * 12 * ITM)
    m.fs.unit.tube_length.fix(53.13 * 12 * ITM)
    m.fs.unit.tube_nrow.fix(20 * 2)
    m.fs.unit.tube_ncol.fix(108)
    m.fs.unit.nrow_inlet.fix(4)
    m.fs.unit.delta_elevation.fix(50)
    m.fs.unit.tube_r_fouling = 0.000176  # (0.001 h-ft^2-F/BTU)
    m.fs.unit.tube_r_fouling = 0.003131  # (0.03131 - 0.1779 h-ft^2-F/BTU)
    if m.fs.unit.config.has_radiation is True:
        m.fs.unit.emissivity_wall.fix(0.7)  # wall emissivity
    # correction factor for overall heat transfer coefficient
    m.fs.unit.fcorrection_htc.fix(1.5)
    # correction factor for pressure drop calc tube side
    m.fs.unit.fcorrection_dp_tube.fix(1.0)
    # correction factor for pressure drop calc shell side
    m.fs.unit.fcorrection_dp_shell.fix(1.0)

    assert degrees_of_freedom(m) == 0
    iscale.calculate_scaling_factors(m)
    m.fs.unit.initialize()

    results = solver.solve(m)
    # Check for optimal solution
    assert check_optimal_termination(results)
    assert value(m.fs.unit.cold_side.properties_out[0].temperature) == pytest.approx(
        tout_1, abs=0.5
    )
    assert value(m.fs.unit.hot_side.properties_out[0].temperature) == pytest.approx(
        tout_2, abs=0.5
    )


def tu(delta_temperature_callback=delta_temperature_underwood_callback):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()
    m.fs.prop_steam = iapws95.Iapws95ParameterBlock()
    m.fs.prop_fluegas = FlueGasParameterBlock()

    m.fs.unit = BoilerHeatExchanger(
        delta_temperature_callback=delta_temperature_callback,
        cold_side={"property_package": m.fs.prop_steam, "has_pressure_change": True},
        hot_side={"property_package": m.fs.prop_fluegas, "has_pressure_change": True},
        has_holdup=False,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        tube_arrangement=TubeArrangement.inLine,
        cold_side_water_phase="Liq",
        has_radiation=True,
    )

    assert_units_consistent(m)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_config_am():
    tc(delta_temperature_amtd_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_config_lm():
    tc(delta_temperature_lmtd_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.unit
def test_config_uw():
    tc(delta_temperature_underwood_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_boiler_hx_am():
    # arithmetic mean is pretty far off
    th(delta_temperature_amtd_callback, tout_1=817.7, tout_2=720)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_boiler_hx_lm():
    th(delta_temperature_lmtd_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_boiler_hx_uw():
    th(delta_temperature_underwood_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_units_am():
    tu(delta_temperature_amtd_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_units_lm():
    tu(delta_temperature_lmtd_callback)


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
@pytest.mark.integration
def test_units_uw():
    tu(delta_temperature_underwood_callback)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Author: Andrew Lee, Alejandro Garciadiego
"""
import pytest
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.unittest import assertStructuredAlmostEqual

from idaes.core import Component, FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core.solvers import get_solver


from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.unit_models import Flash
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE

from idaes.models.properties.modular_properties.examples.CO2_bmimPF6_PR import (
    configuration,
)
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


def _as_quantity(x):
    unit = pyunits.get_units(x)
    if unit is None:
        unit = pyunits.dimensionless
    return value(x) * unit._get_pint_unit()


# Test for configuration dictionaries with parameters from Properties of Gases
# and liquids 4th edition
class TestParamBlock(object):
    @pytest.mark.unit
    def test_build(self):
        model = ConcreteModel()

        model.param = GenericParameterBlock(**configuration)

        assert isinstance(model.param.phase_list, Set)
        assert len(model.param.phase_list) == 2
        for i in model.param.phase_list:
            assert i in ["Liq", "Vap"]
        assert model.param.Liq.is_liquid_phase()
        assert model.param.Vap.is_vapor_phase()

        assert isinstance(model.param.component_list, Set)
        assert len(model.param.component_list) == 2
        for i in model.param.component_list:
            assert i in ["bmimPF6", "carbon_dioxide"]
            assert isinstance(model.param.get_component(i), Component)

        assert isinstance(model.param._phase_component_set, Set)
        assert len(model.param._phase_component_set) == 3
        for i in model.param._phase_component_set:
            assert i in [
                ("Liq", "bmimPF6"),
                ("Liq", "carbon_dioxide"),
                ("Vap", "carbon_dioxide"),
            ]

        assert model.param.config.state_definition == FTPx

        assertStructuredAlmostEqual(
            model.param.config.state_bounds,
            {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (10, 300, 500, pyunits.K),
                "pressure": (5e-4, 1e5, 1e10, pyunits.Pa),
            },
            item_callback=_as_quantity,
        )

        assert model.param.config.phase_equilibrium_state == {("Vap", "Liq"): SmoothVLE}

        assert isinstance(model.param.phase_equilibrium_idx, Set)
        assert len(model.param.phase_equilibrium_idx) == 1
        for i in model.param.phase_equilibrium_idx:
            assert i in ["PE1"]

        assert model.param.phase_equilibrium_list == {
            "PE1": ["carbon_dioxide", ("Vap", "Liq")]
        }

        assert model.param.pressure_ref.value == 101325
        assert model.param.temperature_ref.value == 298.15

        assert model.param.bmimPF6.mw.value == 284.18e-3
        assert model.param.bmimPF6.pressure_crit.value == 24e5
        assert model.param.bmimPF6.temperature_crit.value == 860

        assert model.param.carbon_dioxide.mw.value == 44.010e-3
        assert model.param.carbon_dioxide.pressure_crit.value == 71.8e5
        assert model.param.carbon_dioxide.temperature_crit.value == 304.1

        assert_units_consistent(model)


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()

        model.fs = FlowsheetBlock(dynamic=False)

        model.fs.param = GenericParameterBlock(**configuration)

        model.fs.props = model.fs.param.build_state_block([1], defined_state=True)

        # Fix state
        model.fs.props[1].flow_mol.fix(1)
        model.fs.props[1].temperature.fix(298.15)
        model.fs.props[1].pressure.fix(1214713.75)
        model.fs.props[1].mole_frac_comp["carbon_dioxide"].fix(0.2)
        model.fs.props[1].mole_frac_comp["bmimPF6"].fix(0.8)

        assert degrees_of_freedom(model.fs.props[1]) == 0

        return model

    @pytest.mark.unit
    def test_build(self, model):
        # Check state variable values and bounds
        assert isinstance(model.fs.props[1].flow_mol, Var)
        assert value(model.fs.props[1].flow_mol) == 1
        assert model.fs.props[1].flow_mol.ub == 1000
        assert model.fs.props[1].flow_mol.lb == 0

        assert isinstance(model.fs.props[1].pressure, Var)
        assert value(model.fs.props[1].pressure) == 1214713.75
        assert model.fs.props[1].pressure.ub == 1e10
        assert model.fs.props[1].pressure.lb == 5e-4

        assert isinstance(model.fs.props[1].temperature, Var)
        assert value(model.fs.props[1].temperature) == 298.15
        assert model.fs.props[1].temperature.ub == 500
        assert model.fs.props[1].temperature.lb == 10

        assert isinstance(model.fs.props[1].mole_frac_comp, Var)
        assert len(model.fs.props[1].mole_frac_comp) == 2
        assert value(model.fs.props[1].mole_frac_comp["carbon_dioxide"]) == 0.2
        assert value(model.fs.props[1].mole_frac_comp["bmimPF6"]) == 0.8

        assert_units_consistent(model)

    @pytest.mark.unit
    def test_basic_scaling(self, model):
        model.fs.props[1].calculate_scaling_factors()

        assert len(model.fs.props[1].scaling_factor) == 18
        assert model.fs.props[1].scaling_factor[model.fs.props[1].flow_mol] == 1e-2
        assert (
            model.fs.props[1].scaling_factor[model.fs.props[1].flow_mol_phase["Liq"]]
            == 1e-2
        )
        assert (
            model.fs.props[1].scaling_factor[model.fs.props[1].flow_mol_phase["Vap"]]
            == 1e-2
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].flow_mol_phase_comp["Liq", "bmimPF6"]
            ]
            == 1e-2
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].flow_mol_phase_comp["Liq", "carbon_dioxide"]
            ]
            == 1e-2
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].flow_mol_phase_comp["Vap", "carbon_dioxide"]
            ]
            == 1e-2
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].mole_frac_comp["bmimPF6"]
            ]
            == 1000
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].mole_frac_comp["carbon_dioxide"]
            ]
            == 1000
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].mole_frac_phase_comp["Liq", "bmimPF6"]
            ]
            == 1000
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].mole_frac_phase_comp["Liq", "carbon_dioxide"]
            ]
            == 1000
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].mole_frac_phase_comp["Vap", "carbon_dioxide"]
            ]
            == 1000
        )
        assert model.fs.props[1].scaling_factor[model.fs.props[1].pressure] == 1e-5
        assert model.fs.props[1].scaling_factor[model.fs.props[1].temperature] == 1e-2
        assert (
            model.fs.props[1].scaling_factor[model.fs.props[1]._teq["Vap", "Liq"]]
            == 1e-2
        )
        assert model.fs.props[1].scaling_factor[model.fs.props[1]._t1_Vap_Liq] == 1e-2

        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1]._mole_frac_tbub["Vap", "Liq", "bmimPF6"]
            ]
            == 1
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1]._mole_frac_tbub["Vap", "Liq", "carbon_dioxide"]
            ]
            == 1000
        )
        assert (
            model.fs.props[1].scaling_factor[
                model.fs.props[1].temperature_bubble["Vap", "Liq"]
            ]
            == 1e-2
        )

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.fs.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.fs.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_mol", "mole_frac_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.fs.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Total Molar Flowrate",
                "Total Mole Fraction",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.fs.props[1].report()


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestFlashIntegration(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()

        model.fs = FlowsheetBlock(dynamic=False)

        model.fs.param = GenericParameterBlock(**configuration)

        model.fs.unit = Flash(
            property_package=model.fs.param,
            has_heat_transfer=False,
            has_pressure_change=False,
        )
        # Fix state
        model.fs.unit.inlet.flow_mol.fix(1)
        model.fs.unit.inlet.temperature.fix(200.00)
        model.fs.unit.inlet.pressure.fix(101325)
        model.fs.unit.inlet.mole_frac_comp[0, "carbon_dioxide"].fix(1 / 2)
        model.fs.unit.inlet.mole_frac_comp[0, "bmimPF6"].fix(1 / 2)

        assert degrees_of_freedom(model.fs) == 0

        # Apply scaling - model will not solver without this
        model.fs.unit.control_volume.properties_in[0].calculate_scaling_factors()
        model.fs.unit.control_volume.properties_out[0].calculate_scaling_factors()

        return model

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model, dof=0)

    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.component
    def test_solution(self, model):
        # Check phase equilibrium results
        assert value(
            model.fs.unit.liq_outlet.mole_frac_comp[0, "carbon_dioxide"]
        ) == pytest.approx(0.3119, abs=1e-4)
        assert value(
            model.fs.unit.vap_outlet.mole_frac_comp[0, "carbon_dioxide"]
        ) == pytest.approx(1.0000, abs=1e-4)
        assert value(
            model.fs.unit.vap_outlet.flow_mol[0] / model.fs.unit.liq_outlet.flow_mol[0]
        ) == pytest.approx(0.37619, abs=1e-4)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_report(self, model):
        model.fs.unit.report()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring

__author__ = "Douglas Allan"

import pyomo.environ as pyo
from idaes.core import FlowsheetBlock


def _cell_flowsheet_model(dynamic, time_set, zfaces):
    # function that creates a unit model with cell-level variables for testing
    # subcomponents that require them
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    tset = m.fs.config.time
    znodes = m.fs.znodes = pyo.Set(
        initialize=[(zfaces[i] + zfaces[i + 1]) / 2.0 for i in range(len(zfaces) - 1)]
    )
    iznodes = m.fs.iznodes = pyo.Set(initialize=range(1, len(znodes) + 1))

    m.fs.length_z = pyo.Var(initialize=0.25, units=pyo.units.m)
    m.fs.length_y = pyo.Var(initialize=0.25, units=pyo.units.m)

    m.fs.current_density = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.A / pyo.units.m**2
    )

    m.fs.temperature_z = pyo.Var(tset, iznodes, initialize=1000, units=pyo.units.K)

    m.fs.length_y.fix(0.08)
    m.fs.length_z.fix(0.08)
    m.fs.temperature_z.fix(1000)
    m.fs.current_density.fix(0)

    return m


def _build_test_utility(block, comp_dict, references=None):
    # Takes a unit model and four dictionaries: references (variables),
    # not_references (variables), constraints, and expressions. They should
    # have the attribute name as a key and the length of the attribute as
    # the value. This function goes through and ensures that all these
    # components exist and are the right length, and furthermore they are no
    # unlisted components of these types

    if references is not None:
        for attr in references:
            try:
                comp = getattr(block, attr)
            except AttributeError:
                raise AttributeError(
                    f"Reference {attr} missing from block {block.name}."
                )
            if not comp.is_reference():
                raise AssertionError(
                    f"Attribute {attr} found on block {block.name}, but "
                    "was not Reference."
                )
        for comp in block.component_data_objects(descend_into=False):
            if comp.is_reference():
                if not comp in references:
                    raise AssertionError(
                        f"Unexpected Reference {comp.name} encountered "
                        f"in block {block.name}."
                    )
    for ctype, sub_dict in comp_dict.items():
        for attr, length in sub_dict.items():
            try:
                comp = getattr(block, attr)
            except AttributeError:
                raise AttributeError(f"{ctype} {attr} missing from block {block.name}.")

            if not len(comp) == length:
                raise AssertionError(
                    f"{ctype} {comp.name} was not expected length in block "
                    f"{block.name}."
                )
        for comp in block.component_data_objects(ctype=ctype, descend_into=False):
            short_name = comp.local_name.split("[")[0]
            if not short_name in sub_dict.keys():
                raise AssertionError(
                    f"Unexpected {ctype} {comp.name} encountered in block "
                    f"{block.name}."
                )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains the base class for constructing flowsheet models in the
IDAES modeling framework.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring
from warnings import warn

import pyomo.environ as pe
from pyomo.dae import ContinuousSet
from pyomo.network import Arc
from pyomo.common.config import ConfigValue, ListOf
from pyomo.core.base.units_container import _PyomoUnit

from idaes.core import (
    ProcessBlockData,
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_time_domain,
    DefaultBool,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import DynamicError, ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe

import idaes.logger as idaeslog

# Some more information about this module
__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ["FlowsheetBlock", "FlowsheetBlockData"]

# Set up logger
_log = idaeslog.getLogger(__name__)


class UI:
    """Encapsulate checks for installed 'idaes_ui' dependency.

    Functions exported by this class will run normally if 'idaes_ui' is installed and
    just print warnings if it is not.

    Functions:
        - `visualize(model, model_name, **kwargs)`

    Also has an attribute 'installed' to check directly.
    """

    def __init__(self):
        # pylint: disable=import-outside-toplevel
        try:
            import idaes_ui
        except ImportError:
            idaes_ui = None

        if idaes_ui is None:
            self.visualize = self._visualize_null
            self.installed = False
        else:
            # FIXME the explicit submodule import is needed because the idaes_ui doesn't import its fv submodule
            # otherwise, you get "AttributeError: module 'idaes_ui' has no 'fv' attribute"
            import idaes_ui.fv

            self.visualize = idaes_ui.fv.visualize
            self.installed = True

    def _visualize_null(self, model, model_name, **kwargs):
        self._warn("idaes_ui.fv.visualize")

    @staticmethod
    def _warn(name):
        message = f"Call to {name}() ignored: 'idaes_ui' package is not installed"
        # with stacklevel=3, show the caller of this function's caller
        warn(message, category=RuntimeWarning, stacklevel=3)


@declare_process_block_class(
    "FlowsheetBlock",
    doc="""
    FlowsheetBlock is a specialized Pyomo block for IDAES flowsheet models, and
    contains instances of FlowsheetBlockData.""",
)
class FlowsheetBlockData(ProcessBlockData):
    """
    The FlowsheetBlockData Class forms the base class for all IDAES process
    flowsheet models. The main purpose of this class is to automate the tasks
    common to all flowsheet models and ensure that the necessary attributes of
    a flowsheet model are present.

    The most signfiicant role of the FlowsheetBlockData class is to
    automatically create the time domain for the flowsheet.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic,
**default** - useDefault.
**Valid values:** {
**useDefault** - get flag from parent or False,
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    CONFIG.declare(
        "time",
        ConfigValue(
            default=None,
            domain=is_time_domain,
            description="Flowsheet time domain",
            doc="""Pointer to the time domain for the flowsheet. Users may provide
an existing time domain from another flowsheet, otherwise the flowsheet will
search for a parent with a time domain or create a new time domain and
reference it here.""",
        ),
    )
    CONFIG.declare(
        "time_set",
        ConfigValue(
            default=[0],
            domain=ListOf(float),
            description="Set of points for initializing time domain",
            doc="""Set of points for initializing time domain. This should be a
list of floating point numbers,
**default** - [0].""",
        ),
    )
    CONFIG.declare(
        "time_units",
        ConfigValue(
            description="Units for time domain",
            doc="""Pyomo Units object describing the units of the time domain.
This must be defined for dynamic simulations, default = None.""",
        ),
    )
    CONFIG.declare(
        "default_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Default property package to use in flowsheet",
            doc="""Indicates the default property package to be used by models
within this flowsheet if not otherwise specified,
**default** - None.
**Valid values:** {
**None** - no default property package,
**a ParameterBlock object**.}""",
        ),
    )

    def build(self):
        """
        General build method for FlowsheetBlockData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of flowsheets.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        super(FlowsheetBlockData, self).build()

        self._time_units = None

        # Set up dynamic flag and time domain
        self._setup_dynamics()

    @property
    def time(self):
        # _time will be created by the _setup_dynamics method
        return self._time

    @property
    def time_units(self):
        return self._time_units

    def is_flowsheet(self):
        """
        Method which returns True to indicate that this component is a
        flowsheet.

        Args:
            None

        Returns:
            True
        """
        return True

    def model_check(self):
        """
        This method runs model checks on all unit models in a flowsheet.

        This method searches for objects which inherit from UnitModelBlockData
        and executes the model_check method if it exists.

        Args:
            None

        Returns:
            None
        """
        _log.info("Executing model checks.")
        for o in self.component_objects(descend_into=False):
            if isinstance(o, UnitModelBlockData):
                try:
                    o.model_check()
                except AttributeError:
                    # This should never happen, but just in case
                    _log.warning(
                        "{} Model/block has no model_check method.".format(o.name)
                    )

    def stream_table(self, true_state=False, time_point=0, orient="columns"):
        """
        Method to generate a stream table by iterating over all Arcs in the
        flowsheet.

        Args:
            true_state : whether the state variables (True) or display
                         variables (False, default) from the StateBlocks should
                         be used in the stream table.
            time_point : point in the time domain at which to create stream
                         table (default = 0)
            orient : whether stream should be shown by columns ("columns") or
                     rows ("index")

        Returns:
            A pandas dataframe containing stream table information
        """
        dict_arcs = {}

        for a in self.component_objects(ctype=Arc, descend_into=False):
            dict_arcs[a.local_name] = a

        return create_stream_table_dataframe(
            dict_arcs, time_point=time_point, orient=orient, true_state=true_state
        )

    def visualize(self, model_name, **kwargs) -> "VisualizeResult":
        """
        Starts up a flask server that serializes the model and pops up a
        webpage with the visualization

        Args:
            model_name : The name of the model

        Keyword Args:
            **kwargs: Additional keywords for :func:`idaes.core.ui.fv.visualize()`

        Returns:
            The :class:`idaes_ui.fv.fsvis.VisualizeResult` instance returned by :meth:`UI.visualize`
        """
        visualize_result = UI().visualize(self, model_name, **kwargs)
        return visualize_result

    def _get_stream_table_contents(self, time_point=0):
        """
        Calls stream_table method and returns result
        """
        return self.stream_table(time_point)

    def _setup_dynamics(self):
        # Look for parent flowsheet
        fs = self.flowsheet()

        # Check the dynamic flag, and retrieve if necessary
        if self.config.dynamic == useDefault:
            if fs is None:
                # No parent, so default to steady-state and warn user
                _log.warning(
                    "{} is a top level flowsheet, but dynamic flag "
                    "set to useDefault. Dynamic "
                    "flag set to False by default".format(self.name)
                )
                self.config.dynamic = False

            else:
                # Get dynamic flag from parent flowsheet
                self.config.dynamic = fs.config.dynamic

        # Check for case when dynamic=True, but parent dynamic=False
        elif self.config.dynamic is True:
            if fs is not None and fs.config.dynamic is False:
                raise DynamicError(
                    "{} trying to declare a dynamic model within "
                    "a steady-state flowsheet. This is not "
                    "supported by the IDAES framework. Try "
                    "creating a dynamic flowsheet instead, and "
                    "declaring some models as steady-state.".format(self.name)
                )

        # Validate units for time domain
        if self.config.time is None and fs is not None:
            # We will get units from parent
            pass
        elif self.config.time_units is None and self.config.dynamic:
            raise ConfigurationError(
                f"{self.name} - no units were specified for the time domain. "
                f"Units must be specified for dynamic models."
            )
        elif self.config.time_units is None and not self.config.dynamic:
            _log.debug("No units specified for steady-state time domain.")
        elif not isinstance(self.config.time_units, _PyomoUnit):
            raise ConfigurationError(
                "{} unrecognised value for time_units argument. This must be "
                "a Pyomo Unit object (not a compound unit).".format(self.name)
            )

        if self.config.time is not None:
            # Validate user provided time domain
            if self.config.dynamic is True and not isinstance(
                self.config.time, ContinuousSet
            ):
                raise DynamicError(
                    "{} was set as a dynamic flowsheet, but time domain "
                    "provided was not a ContinuousSet.".format(self.name)
                )
            add_object_reference(self, "_time", self.config.time)
            self._time_units = self.config.time_units
        else:
            # If no parent flowsheet, set up time domain
            if fs is None:
                # Create time domain
                if self.config.dynamic:
                    # Check if time_set has at least two points
                    if len(self.config.time_set) < 2:
                        # Check if time_set is at default value
                        if self.config.time_set == [0.0]:
                            # If default, set default end point to be 1.0
                            self.config.time_set = [0.0, 1.0]
                        else:
                            # Invalid user input, raise Exception
                            raise DynamicError(
                                "Flowsheet provided with invalid "
                                "time_set attribute - must have at "
                                "least two values (start and end)."
                            )
                    # For dynamics, need a ContinuousSet
                    self._time = ContinuousSet(initialize=self.config.time_set)
                else:
                    # For steady-state, use an ordered Set
                    self._time = pe.Set(initialize=self.config.time_set, ordered=True)
                self._time_units = self.config.time_units

                # Set time config argument as reference to time domain
                self.config.time = self._time
            else:
                # Set time config argument to parent time
                self.config.time = fs.time
                add_object_reference(self, "_time", fs.time)
                # We control time units
                # pylint: disable-next=protected-access
                self._time_units = fs._time_units



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for tray column unit model (single feed tray, no side draws).

Author: Jaffer Ghouse
"""
import pytest
from pyomo.environ import check_optimal_termination, ConcreteModel, value
from pyomo.util.check_units import assert_units_consistent
import pyomo.common.unittest as unittest

from idaes.core import FlowsheetBlock
from idaes.models_extra.column_models import TrayColumn
from idaes.models_extra.column_models.condenser import CondenserType, TemperatureSpec
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.performance import PerformanceBaseClass

from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)

from idaes.models.properties.modular_properties.examples.BT_ideal import configuration
import idaes.core.util.scaling as iscale

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


@pytest.mark.unit
def test_config():

    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = TrayColumn(
        number_of_trays=10,
        feed_tray_location=5,
        condenser_type=CondenserType.totalCondenser,
        condenser_temperature_spec=TemperatureSpec.customTemperature,
        property_package=m.fs.properties,
        has_heat_transfer=False,
        has_pressure_change=False,
    )

    assert len(m.fs.unit.config) == 12

    assert m.fs.unit.feed_tray.config.is_feed_tray

    assert hasattr(m.fs.unit, "condenser")
    assert hasattr(m.fs.unit, "reboiler")

    assert hasattr(m.fs.unit, "rectification_section")
    assert hasattr(m.fs.unit, "stripping_section")


def build_model_btx_ftpz():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = BTXParameterBlock(
        valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal"
    )

    m.fs.unit = TrayColumn(
        number_of_trays=10,
        feed_tray_location=5,
        condenser_type=CondenserType.totalCondenser,
        condenser_temperature_spec=TemperatureSpec.atBubblePoint,
        property_package=m.fs.properties,
        has_heat_transfer=False,
        has_pressure_change=False,
    )

    # Inlet feed conditions
    m.fs.unit.feed.flow_mol.fix(40)
    m.fs.unit.feed.temperature.fix(368)
    m.fs.unit.feed.pressure.fix(101325)
    m.fs.unit.feed.mole_frac_comp[0, "benzene"].fix(0.5)
    m.fs.unit.feed.mole_frac_comp[0, "toluene"].fix(0.5)

    # unit level inputs
    m.fs.unit.condenser.reflux_ratio.fix(1.4)
    m.fs.unit.condenser.condenser_pressure.fix(101325)

    m.fs.unit.reboiler.boilup_ratio.fix(1.3)

    # iscale.calculate_scaling_factors(m)

    return m


@pytest.mark.performance
class Test_TrayColumn_Performance(PerformanceBaseClass, unittest.TestCase):
    def build_model(self):
        return build_model_btx_ftpz()

    def initialize_model(self, model):
        model.fs.unit.initialize()


class TestBTXIdealFcTP:
    @pytest.fixture(scope="class")
    def btx_fctp(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = BTXParameterBlock(
            valid_phase=("Liq", "Vap"), activity_coeff_model="Ideal", state_vars="FcTP"
        )

        m.fs.unit = TrayColumn(
            number_of_trays=10,
            feed_tray_location=5,
            condenser_type=CondenserType.totalCondenser,
            condenser_temperature_spec=TemperatureSpec.atBubblePoint,
            property_package=m.fs.properties,
            has_heat_transfer=False,
            has_pressure_change=False,
        )

        # Inlet feed conditions
        m.fs.unit.feed.flow_mol_comp[0, "benzene"].fix(20)
        m.fs.unit.feed.flow_mol_comp[0, "toluene"].fix(20)
        m.fs.unit.feed.temperature.fix(368)
        m.fs.unit.feed.pressure.fix(101325)

        # unit level inputs
        m.fs.unit.condenser.reflux_ratio.fix(1.4)
        m.fs.unit.condenser.condenser_pressure.fix(101325)

        m.fs.unit.reboiler.boilup_ratio.fix(1.3)

        # iscale.calculate_scaling_factors(m)

        return m

    @pytest.mark.unit
    def test_build(self, btx_fctp):
        assert len(btx_fctp.fs.unit.config) == 12

        assert btx_fctp.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_fctp.fs.unit, "condenser")
        assert hasattr(btx_fctp.fs.unit, "reboiler")

        assert hasattr(btx_fctp.fs.unit, "rectification_section")
        assert hasattr(btx_fctp.fs.unit, "stripping_section")

    @pytest.mark.unit
    def test_dof(self, btx_fctp):
        assert degrees_of_freedom(btx_fctp.fs.unit) == 0

    @pytest.mark.component
    def test_units(self, btx_fctp):
        assert_units_consistent(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_fctp):
        initialization_tester(btx_fctp)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_fctp):
        results = solver.solve(btx_fctp)

        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_fctp):

        # Distillate port - btx_fctp
        assert pytest.approx(16.856, rel=1e-2) == value(
            btx_fctp.fs.unit.condenser.distillate.flow_mol_comp[0, "benzene"]
        )
        assert pytest.approx(2.121, rel=1e-2) == value(
            btx_fctp.fs.unit.condenser.distillate.flow_mol_comp[0, "toluene"]
        )
        assert pytest.approx(355.642, rel=1e-4) == value(
            btx_fctp.fs.unit.condenser.distillate.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_fctp.fs.unit.condenser.distillate.pressure[0]
        )

        # Bottoms port - btx_fctp
        assert pytest.approx(3.143, rel=1e-2) == value(
            btx_fctp.fs.unit.reboiler.bottoms.flow_mol_comp[0, "benzene"]
        )
        assert pytest.approx(17.876, rel=1e-3) == value(
            btx_fctp.fs.unit.reboiler.bottoms.flow_mol_comp[0, "toluene"]
        )
        assert pytest.approx(377.33, rel=1e-4) == value(
            btx_fctp.fs.unit.reboiler.bottoms.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_fctp.fs.unit.reboiler.bottoms.pressure[0]
        )


class TestBTXIdealFTPz:
    @pytest.fixture(scope="class")
    def btx_ftpz(self):
        return build_model_btx_ftpz()

    @pytest.mark.unit
    def test_build(self, btx_ftpz):
        assert len(btx_ftpz.fs.unit.config) == 12

        assert btx_ftpz.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_ftpz.fs.unit, "condenser")
        assert hasattr(btx_ftpz.fs.unit, "reboiler")

        assert hasattr(btx_ftpz.fs.unit, "rectification_section")
        assert hasattr(btx_ftpz.fs.unit, "stripping_section")

    @pytest.mark.unit
    def test_dof(self, btx_ftpz):
        assert degrees_of_freedom(btx_ftpz.fs.unit) == 0

    @pytest.mark.component
    def test_units(self, btx_ftpz):
        assert_units_consistent(btx_ftpz)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_ftpz):
        initialization_tester(btx_ftpz)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_ftpz):
        results = solver.solve(btx_ftpz)

        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_ftpz):

        # Distillate port - btx_ftpz
        assert pytest.approx(18.978, rel=1e-2) == value(
            btx_ftpz.fs.unit.condenser.distillate.flow_mol[0]
        )
        assert pytest.approx(0.8882, rel=1e-2) == value(
            btx_ftpz.fs.unit.condenser.distillate.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.1118, rel=1e-2) == value(
            btx_ftpz.fs.unit.condenser.distillate.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(355.642, rel=1e-4) == value(
            btx_ftpz.fs.unit.condenser.distillate.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz.fs.unit.condenser.distillate.pressure[0]
        )

        # Bottoms port - btx_ftpz
        assert pytest.approx(21.021, rel=1e-2) == value(
            btx_ftpz.fs.unit.reboiler.bottoms.flow_mol[0]
        )
        assert pytest.approx(0.1495, rel=1e-2) == value(
            btx_ftpz.fs.unit.reboiler.bottoms.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.8504, rel=1e-3) == value(
            btx_ftpz.fs.unit.reboiler.bottoms.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(377.337, rel=1e-4) == value(
            btx_ftpz.fs.unit.reboiler.bottoms.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz.fs.unit.reboiler.bottoms.pressure[0]
        )


class TestBTXIdealGeneric:
    @pytest.fixture(scope="class")
    def btx_ftpz_generic(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.properties = GenericParameterBlock(**configuration)

        m.fs.unit = TrayColumn(
            number_of_trays=10,
            feed_tray_location=5,
            condenser_type=CondenserType.totalCondenser,
            condenser_temperature_spec=TemperatureSpec.atBubblePoint,
            property_package=m.fs.properties,
            has_heat_transfer=False,
            has_pressure_change=False,
        )

        # Inlet feed conditions
        m.fs.unit.feed.flow_mol.fix(100)
        m.fs.unit.feed.temperature.fix(368)
        m.fs.unit.feed.pressure.fix(101325)
        m.fs.unit.feed.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.feed.mole_frac_comp[0, "toluene"].fix(0.5)

        # unit level inputs
        m.fs.unit.condenser.reflux_ratio.fix(1.4)
        m.fs.unit.condenser.condenser_pressure.fix(101325)

        m.fs.unit.reboiler.boilup_ratio.fix(1.3)

        return m

    @pytest.mark.unit
    def test_build(self, btx_ftpz_generic):
        assert len(btx_ftpz_generic.fs.unit.config) == 12

        assert btx_ftpz_generic.fs.unit.feed_tray.config.is_feed_tray

        assert hasattr(btx_ftpz_generic.fs.unit, "condenser")
        assert hasattr(btx_ftpz_generic.fs.unit, "reboiler")

        assert hasattr(btx_ftpz_generic.fs.unit, "rectification_section")
        assert hasattr(btx_ftpz_generic.fs.unit, "stripping_section")

        assert len(btx_ftpz_generic.fs.unit.config) == 12

    @pytest.mark.unit
    def test_dof(self, btx_ftpz_generic):
        assert degrees_of_freedom(btx_ftpz_generic) == 0

    @pytest.mark.component
    def test_units(self, btx_ftpz_generic):
        assert_units_consistent(btx_ftpz_generic)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx_ftpz_generic):
        initialization_tester(btx_ftpz_generic)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx_ftpz_generic):
        results = solver.solve(btx_ftpz_generic)

        assert check_optimal_termination(results)

    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx_ftpz_generic):

        # Distillate port - btx_ftpz
        assert pytest.approx(49.28, rel=1e-2) == value(
            btx_ftpz_generic.fs.unit.condenser.distillate.flow_mol[0]
        )
        assert pytest.approx(0.8784, rel=1e-2) == value(
            btx_ftpz_generic.fs.unit.condenser.distillate.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.1215, rel=1e-2) == value(
            btx_ftpz_generic.fs.unit.condenser.distillate.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(355.853, rel=1e-4) == value(
            btx_ftpz_generic.fs.unit.condenser.distillate.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz_generic.fs.unit.condenser.distillate.pressure[0]
        )

        # Bottoms port - btx_ftpz
        assert pytest.approx(50.711, rel=1e-2) == value(
            btx_ftpz_generic.fs.unit.reboiler.bottoms.flow_mol[0]
        )
        assert pytest.approx(0.1322, rel=1e-2) == value(
            btx_ftpz_generic.fs.unit.reboiler.bottoms.mole_frac_comp[0, "benzene"]
        )
        assert pytest.approx(0.8677, rel=1e-3) == value(
            btx_ftpz_generic.fs.unit.reboiler.bottoms.mole_frac_comp[0, "toluene"]
        )
        assert pytest.approx(378.0433, rel=1e-4) == value(
            btx_ftpz_generic.fs.unit.reboiler.bottoms.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx_ftpz_generic.fs.unit.reboiler.bottoms.pressure[0]
        )



# -*- coding: UTF-8 -*-
#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains utility functions for mathematical operators of use in
equation oriented models.
"""

from pyomo.environ import Param, log, sqrt

__author__ = "Andrew Lee"


def smooth_abs(a, eps=1e-4):
    """General function for creating an expression for a smooth minimum or
    maximum.

    .. math:: |a| = sqrt(a^2 + eps^2)

    Args:
        a : term to get absolute value from (Pyomo component, float or int)
        eps : smoothing parameter (Param, float or int) (default=1e-4)

    Returns:
        An expression for the smoothed absolute value operation.
    """
    # Check type of eps
    if not isinstance(eps, (float, int, Param)):
        raise TypeError("smooth_abs eps argument must be a float, int or Pyomo Param")

    # Create expression
    try:
        expr = (a**2 + eps**2) ** 0.5
    except TypeError:
        raise TypeError(
            "Unsupported argument type for smooth_abs. Must be "
            "a Pyomo Var, Param or Expression, or a float or int."
        )

    return expr


def smooth_minmax(a, b, eps=1e-4, sense="max"):
    """General function for creating an expression for a smooth minimum or
    maximum. Uses the smooth_abs operator.

    .. math:: minmax(a, b) = 0.5*(a+b +- |a-b|)

    Args:
        a : first term in mix or max function (Pyomo component, float or int)
        b : second term in min or max function (Pyomo component, float or int)
        eps : smoothing parameter (Param, float or int) (default=1e-4)
        sense : 'mim' or 'max' (default = 'max')

    Returns:
        An expression for the smoothed minimum or maximum operation.
    """
    # Check type of eps
    if not isinstance(eps, (float, int, Param)):
        raise TypeError(
            "Smooth {} eps argument must be a float, int or "
            "Pyomo Param".format(sense)
        )

    # Set sense of expression
    if sense == "max":
        mm = 1
    elif sense == "min":
        mm = -1
    else:
        raise ValueError(
            "Unrecognised sense argument to smooth_minmax. Must be 'min' or 'max'."
        )

    # Create expression
    try:
        expr = 0.5 * (a + b + mm * smooth_abs(a - b, eps))
    except TypeError:
        raise TypeError(
            "Unsupported argument type for smooth_{}. Must be "
            "a Pyomo Var, Param or Expression, or a float or int.".format(sense)
        )

    return expr


def smooth_max(a, b, eps=1e-4):
    """Smooth maximum operator, using smooth_abs operator.

    .. math:: max(a, b) = 0.5*(a+b + |a-b|)

    Args:
        a : first term in max function
        b : second term in max function
        eps : smoothing parameter (Param or float, default = 1e-4)

    Returns:
        An expression for the smoothed maximum operation.
    """
    expr = smooth_minmax(a, b, eps, sense="max")
    return expr


def smooth_min(a, b, eps=1e-4):
    """Smooth minimum operator, using smooth_abs operator.

    .. math:: max(a, b) = 0.5*(a+b - |a-b|)

    Args:
        a : first term in min function
        b : second term in min function
        eps : smoothing parameter (Param or float, default = 1e-4)

    Returns:
        An expression for the smoothed minimum operation.
    """
    expr = smooth_minmax(a, b, eps, sense="min")
    return expr


def smooth_bound(val, lb, ub, eps=1e-4, eps_lb=None, eps_ub=None):
    """Returns a smooth expression that returns approximately the value of val
    between lb and ub, the value of ub if val > ub and the value of lb if
    val < lb. The expression returned is

    ..math:: smooth_min(smooth_max(val, lb, eps_lb), ub, eps_ub)

    **Example Usage** This function is useful when calculating the value of a
    quantity with physical limitations that may not be otherwise captured. A
    good example of this is using a controller to manipulate a valve to maintain,
    for example, a set flow rate. Under some process conditions the controller
    may indicate the valve should be more than fully opened or closed. In this
    example, the smooth_bound expression can be applied to the unbounded
    controller output to get a bounded controller output, ensuring the valve
    opening stays in the feasible region, while maintaining smoothness.

    Args:
        val: an expression for a bounded version of this value is returned
        lb: lower bound
        ub: lower bound
        eps: smoothing parameter
        eps_lb: (optional) lower bound smoothing parameter if not provided use eps
        eps_ub: (optional) upper bound smoothing parameter if not provided use eps

    Returns:
        expression: smooth bounded version of val.
    """
    if eps_lb is None:
        eps_lb = eps
    if eps_ub is None:
        eps_ub = eps
    return smooth_min(smooth_max(val, lb, eps_lb), ub, eps_ub)


def safe_sqrt(a, eps=1e-4):
    """Returns the square root of max(a, 0) using the smooth_max expression.
    This can be used to avoid transient evaluation errors when changing a model
    from one state to another.  This can be used when a at the solution is not
    expected to be near 0.

    Args:
        a: Pyomo expression
        eps: epsilon parameter for smooth max

    Returns:
        approximately sqrt(max(a, 0))
    """
    return sqrt(smooth_max(a, 0, eps))


def safe_log(a, eps=1e-4):
    """Returns the log of max(a, eps) using the smooth_max expression.
    This can be used to avoid transient evaluation errors when changing a model
    from one state to another.  This can be used when at the solution, a >> eps.

    Args:
        a: Pyomo expression
        eps: epsilon parameter for smooth max

    Returns:
        approximately log(max(a, eps))
    """
    return log(smooth_max(a, eps, eps=eps))
